import skimage.morphology

from enstools.feature.identification import IdentificationStrategy
import xarray as xr
import numpy as np
import os, sys
from .util import compute_field, interpolate_wavetroughs, create_wt_troughs_and_areas, bpf
from .plotting import plot_kw_style

import json
from types import SimpleNamespace
from matplotlib import pyplot as plt
from enstools.feature.identification._proto_gen import african_easterly_waves_pb2
import cartopy.crs as ccrs

from scipy.signal import find_peaks
from scipy.ndimage import maximum_filter, generate_binary_structure
from skimage.draw import line_aa
from enstools.feature.util.enstools_utils import get_u_var, get_v_var, get_vertical_dim, get_longitude_dim, get_latitude_dim, get_init_time_dim, get_valid_time_dim
import threading
from skimage.draw import line
from IMTreatment.vortex_criterions import get_lambda2, get_gamma
from IMTreatment.core import VectorField, ScalarField, SpatialScalarFields, SpatialVectorFields

class AEWVortexIdentification(IdentificationStrategy):

    def __init__(self, **kwargs):
        """
        Initialize

        """

        import enstools.feature.identification.aew_vortices.configuration as cfg
        self.config = cfg  # config
        self.field = self.config.field
        self.bpf = self.config.bpf
        
        json_path = cfg.in_wts
        with open(json_path) as f:
            wt_data = json.load(f, object_hook=lambda x: SimpleNamespace(**x))

        self.json_desc = wt_data
        self.circles = xr.load_dataset(self.config.circles_file).circles

        pass

    def precompute(self, dataset: xr.Dataset, **kwargs):
        print("Precompute for AEW identification...")

        plt.switch_backend('agg')  # this is thread safe matplotlib but cant display.

        # --------------- CLIMATOLOGY

        u_name = self.config.u_dim if hasattr(self.config, 'u_dim') and (self.config.u_dim is not None) else get_u_var(dataset)
        v_name = self.config.v_dim if hasattr(self.config, 'v_dim') and (self.config.v_dim is not None) else get_v_var(dataset)

        if u_name is None or v_name is None:
            print("Could not locate u and v fields in dataset. Needed to compute advection terms.")
            exit()

        level_str = get_vertical_dim(dataset)
        lat_str = get_latitude_dim(dataset)
        lon_str = get_longitude_dim(dataset)
        init_time_str = get_init_time_dim(dataset)
        valid_time_str = get_valid_time_dim(dataset)

        # --------------- SUBSET DATA ACCORDING TO CFG
        start_date_dt = np.datetime64(self.config.start_date) if self.config.start_date is not None else None
        end_date_dt = np.datetime64(self.config.end_date) if self.config.end_date is not None else None

        # if data is lon=0..360, change it to -180..180
        dataset.coords[lon_str] = (dataset.coords[lon_str] + 180) % 360 - 180
        dataset = dataset.sortby(dataset[lon_str])

        sel_levels = self.config.levels
        filter_time_str = init_time_str if init_time_str is not None else valid_time_str

        # subset lon,lat,lev,time
        dataset = dataset.sel(
            # **{lat_str: slice(lat_range[0], lat_range[1])},
            # **{lon_str: slice(lon_range[0], lon_range[1])},
            **{level_str: sel_levels},
            **{filter_time_str: slice(start_date_dt, end_date_dt)})

        # get the data we want to investigate
        dataset = dataset.sortby(lat_str) # in case of descending

        if len(dataset.time.values) == 0:
            print("Given start and end time leads to no data to process.")
            exit(1)

        # --------------- DO NUMPY PARALLELIZED STUFF: CREATE TROUGH MASKS

        # BPF (optional) TODO
        if self.bpf:
            dataset = bpf(dataset, u_name, v_name, min_days=2, max_days=6)
            self.config.u_dim = 'u_bpf'
            self.config.v_dim = 'v_bpf'
            u_name = 'u_bpf'
            v_name = 'v_bpf'

        self.json_desc = interpolate_wavetroughs(self.json_desc, african_easterly_waves_pb2, self.pb_dataset)

        # create field to use
        dataset = compute_field(dataset, u_name, v_name, self.field)
        dataset = dataset.transpose(..., lat_str, lon_str)

        data_field = dataset[self.field]
        dataset['vortices'] = xr.zeros_like(dataset[self.field], dtype=int)

        """
        l2 = xr.zeros_like(dataset[u_name], dtype=float)
        for time in dataset.time.values:
            print(time)
            dt = dataset.sel(level=850, time=time)

            sx = ScalarField()
            sx.import_from_arrays(dt.latitude.values.tolist(), dt.longitude.values.tolist(), dt[u_name].data)
            sy = ScalarField()
            sy.import_from_arrays(dt.latitude.values.tolist(), dt.longitude.values.tolist(), dt[v_name].data)

            v = VectorField()
            v.import_from_sfs(sx, sy)

            cur_l2 = get_gamma(v)
            print(cur_l2)
            l2.loc[dict(level=850, time=time)] = cur_l2.values
        print(l2)
        dataset['lambda2'] = l2
        dataset['lambda2'].attrs['long_name'] = 'lambda2'
        dataset['lambda2'].attrs['standard_name'] = 'l2'
        dataset['lambda2'].attrs['units'] = 'dunno'
        """

        dataset = create_wt_troughs_and_areas(dataset, self.json_desc, self.circles, False, d=self.config.radius)

        # search local max.
        neighborhood = generate_binary_structure(2, 2)
        # TODO depending on num dimensions. here 4d, with latlon back.
        neighborhood = np.pad([[neighborhood]], ((1,1),(1,1),(0,0),(0,0)))
        max_mask = maximum_filter(data_field.values, footprint=neighborhood) == data_field.values

        # find peaks can give better results
        # max_lat = find_peaks(data_field.values, np.greater, axis=-2) # if more sophisiticated:
        # max_lon = find_peaks(data_field.values, np.greater, axis=-1) #  scipy.signal.find_peaks

        dataset['vortices'].data = max_mask

        # out: set data to 1, fakes to 2
        dataset['wt_rea'] = xr.where(dataset.wt_rea > 0, 1, dataset.wt_rea)
        dataset['wt_rea'] = xr.where(dataset.wt_rea < 0, 2, dataset.wt_rea)
        dataset['infl_rea'] = xr.where(dataset.infl_rea > 0, 1, dataset.infl_rea)
        dataset['infl_rea'] = xr.where(dataset.infl_rea < 0, 2, dataset.infl_rea)

        # FILTER by threshold given in config.
        thr = self.config.base_threshold[self.field]
        if self.bpf:
            thr = thr / 2 # TODO?
        dataset['vortices'] = xr.where(dataset[self.field] > thr, dataset.vortices, 0)
        # FILTER by area -> in vicinity of WT
        dataset['vortices'] = xr.where(dataset['infl_rea'] != 0, dataset.vortices, 0)
        dataset['vortices'] = dataset.vortices.transpose(..., lat_str, lon_str)


        # self.config.prominence_radius



        # filter vortices somehow, before? prob need after
        # add them to WT descs.

        # filter max points by WT mask
        # interpolate_wave_areas() # for gaps in tracks. -> also add json info? maybe create WT area ourself.
        #     for each child in track node: if dt > 6h: add inbetween "fake" wt to json
        # put lows in.
        # maybe depending on whether they form a track?


        #if self.config.cv_name not in dataset.data_vars:
        #    print("Curvature Vorticity not found, trying to compute it out of U and V...")
        #    dataset = compute_cv(dataset, u_name, v_name, self.config.cv_name)

        # make dataset to 2.5 (or same as cv_clim)
        # dataset = dataset.interp({lat_str: cv_clim.coords[lat_str], lon_str: cv_clim.coords[lon_str]})

        # make sure that lat and lon are last two dimensions
        # dataset = dataset.transpose(..., lat_str, lon_str)


        # dataset['wavetroughs'] = wt

        return dataset

    def identify(self, data_chunk: xr.Dataset, **kwargs):

        objs = []
        v = data_chunk.vortices

        ds = data_chunk.stack(z=["latitude", "longitude"])
        ds = ds.where(ds.vortices > 0, drop=True)

        # TODO maybe create WT+area also in here, so we have direct reference to IDs
        # CLUSTER
        # prominence of 2deg -> keep if it is the max peak in a 2deg circle?
        #                    -> or discard if any peak within 2deg is higher? or is this the same?
        # For each vortex...
        for z in range(len(ds.z.data)):
            if ds.vortices[z].item() == 0:
                print("skip?")
                continue
            lon_vortex = ds.longitude[z].item()
            lat_vortex = ds.latitude[z].item()
            data_vertex = ds[self.field][z].item()

            circle = self.circles.sel(longitude_center=lon_vortex, latitude_center=lat_vortex)

            # is this efficient? compare manually with all other vortex pts
            for zc in range(len(ds.z.data)):
                lon_vc = ds.longitude[zc].item()
                lat_vc = ds.latitude[zc].item()
                data_c = ds[self.field][zc].item()

                distance_c = circle.sel(longitude=lon_vc, latitude=lat_vc).item()
                # if distance in promenance radius and other point has higher vorticity -> delete this point.
                if distance_c < self.config.prominence_radius and data_c > data_vertex: # TODO what happens if two with exactly the same cv?
                    # vortex in vicinity is higher. discard this.
                    v.loc[dict(longitude=lon_vortex, latitude=lat_vortex)] = 0
                    break

        # data_chunk.vortices.values = skimage.morphology.binary_dilation(data_chunk.vortices.values)

        # for each vortex
        #  associate with its corresponding WT
        #  put into json



        """
        id_ = 1
        for path in []:

            # get new object, set id
            # o = self.get_new_object()
            # o.id = id_
            # populate it
            # populate_object(o.properties, path, self.config)

            # add to objects if keep
            # if not self.keep_wavetrough(o.properties):
            #     continue

            # objs.append(o)
            # id_ += 1
            break
        """

        return data_chunk, objs

    def postprocess(self, dataset: xr.Dataset, data_desc, **kwargs):


        plot_kw_style(dataset, self.json_desc, self.config)
        print(dataset)
        dataset.to_netcdf("/home/ws/he7273/phd_all/data/coll_oper/jjaso2021_test.nc")
        exit()
        # add vortices to tracks?



        return dataset, data_desc

